import functools
import random
import sys
from collections import deque
from itertools import chain
from types import FunctionType, ModuleType

import matplotlib.pyplot as plt
import seaborn as sns

from effidict import PickleDict

sns.set()

# Set of types that signify "atomic" non-iterable types
BLACKLIST = type, ModuleType, FunctionType


def generate_random_sequence(length):
    """Generates a random DNA sequence of a given length."""
    return "".join(random.choice("ACGT") for _ in range(length))


def total_size(o, handlers={}):
    """Returns the approximate memory footprint of an object and all of its contents."""

    # Handler for standard dictionaries
    dict_handler = lambda d: chain.from_iterable(d.items())

    # Handler for LRUDict
    lrudict_handler = lambda d: chain.from_iterable(d.items())

    # Define all default handlers
    all_handlers = {
        tuple: iter,
        list: iter,
        deque: iter,
        dict: dict_handler,
        set: iter,
        frozenset: iter,
        PickleDict: lrudict_handler,
    }

    all_handlers.update(handlers)  # user handlers take precedence

    seen = set()  # track which object id's have already been seen
    default_size = sys.getsizeof(0)  # estimate sizeof object without __sizeof__

    def sizeof(obj):
        if id(obj) in seen:  # do not double count the same object
            return 0
        seen.add(id(obj))
        s = sys.getsizeof(obj, default_size)

        for typ, handler in all_handlers.items():
            if isinstance(obj, typ):
                s += sum(map(sizeof, handler(obj)))
                break
        return s

    return sizeof(o)


def memory_size_decorator(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        size = total_size(result) * 10**-6
        # print(f"Total memory size of the object returned by '{func.__name__}': {size} bytes")
        return result, size

    return wrapper


def plot_times_and_sizes(times, sizes, n_regions_values):
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(20, 4))

    ax1.plot(n_regions_values, sizes, marker="o")
    ax1.set_title("Size of Dictionary on Memory vs. number of regions")
    ax1.set_xlabel("Number of regions")
    ax1.set_ylabel("Size on Memory (Mega bytes)")

    ax2.plot(n_regions_values, times, marker="o")
    ax2.set_title("Execution time vs. number of regions")
    ax2.set_xlabel("Number of regions")
    ax2.set_ylabel("Time (s)")

    fig.show()


def plot_times(all_times, n_regions_values):
    plt.figure(figsize=(10, 4))
    for times, legend in all_times:
        plt.plot(n_regions_values, times, marker="o", label=legend)
    plt.title("Execution time vs. number of regions")
    plt.xlabel("Number of regions")
    plt.ylabel("Time (s)")
    plt.grid(True)
    plt.legend()
    plt.show()



import time
from abc import abstractmethod
import sqlite3  
import json
import os
import pickle
import shutil

try:
    import h5py
    import numpy as np
except Exception:
    h5py = None
    np = None

class DiskBackend:
    def __init__(self, storage_path):
        self.storage_path = storage_path + f"{int(time.time())}_{id(self)}"
        
    @abstractmethod
    def serialize(self, key, value):
        pass

    @abstractmethod
    def deserialize(self, key):
        pass

    @abstractmethod
    def del_item(self, key):
        pass

    @abstractmethod
    def keys(self):
        pass

    @abstractmethod
    def destroy(self):
        pass

    def load_from_dict(self, dictionary):
        for key, value in dictionary.items():
            self.serialize(key, value)


class SqliteBackend(DiskBackend):
    def __init__(self, storage_path):
        super().__init__(storage_path)
        self.conn = sqlite3.connect(self.storage_path)
        self.cursor = self.conn.cursor()
        self.cursor.execute(
            "CREATE TABLE IF NOT EXISTS data (key TEXT PRIMARY KEY, value TEXT)"
        )

    def serialize(self, key, value):
        json_value = json.dumps(value)
        self.cursor.execute(
            "REPLACE INTO data (key, value) VALUES (?, ?)", (key, json_value)
        )
        self.conn.commit()

    def deserialize(self, key):
        self.cursor.execute("SELECT value FROM data WHERE key=?", (key,))
        result = self.cursor.fetchone()
        if result:
            return json.loads(result[0])
        raise KeyError(key)
    
    def del_item(self, key):
        self.cursor.execute("DELETE FROM data WHERE key=?", (key,))
        self.conn.commit()

    def keys(self):
        self.cursor.execute("SELECT key FROM data")
        return [key[0] for key in self.cursor.fetchall()]
    
    def load_from_dict(self, dictionary):
        with self.conn:
            items_to_insert = [
                (key, json.dumps(value)) for key, value in dictionary.items()
            ]
            self.cursor.executemany(
                "REPLACE INTO data (key, value) VALUES (?, ?)",
                items_to_insert,
            )

    def destroy(self):
        self.conn.close()
        os.remove(self.storage_path)


class PickleBackend(DiskBackend):
    def __init__(self, storage_path):
        super().__init__(storage_path)
        os.makedirs(self.storage_path, exist_ok=True)

    def serialize(self, key, value):
        with open(os.path.join(self.storage_path, key), "wb") as f:
            pickle.dump(value, f)

    def deserialize(self, key):
        try:
            with open(os.path.join(self.storage_path, key), "rb") as f:
                return pickle.load(f)
        except FileNotFoundError:
            raise KeyError(key)
        
    def del_item(self, key):
        try:
            os.remove(os.path.join(self.storage_path, key))
        except FileNotFoundError:
            pass

    def keys(self):
        return os.listdir(self.storage_path)

    def destroy(self):
        shutil.rmtree(self.storage_path)


class Hdf5Backend(DiskBackend):
    def __init__(self, storage_path):
        if h5py is None or np is None:
            raise ImportError("h5py and numpy are required for Hdf5Backend but are not installed.")
        super().__init__(storage_path)
        self.file = h5py.File(self.storage_path, "w")

    def serialize(self, key, value):
        """
        Store data and its Python type as an HDF5 attribute.
        """
        if key in self.file:
            del self.file[key] 

        type_name = type(value).__name__
        
        if value is None:
            ds = self.file.create_dataset(key, data=h5py.Empty("f"))
        elif isinstance(value, (dict, list, tuple)):
            ds = self.file.create_dataset(key, data=repr(value))
        else:
            ds = self.file.create_dataset(key, data=value)
            
        ds.attrs['python_type'] = type_name

    def deserialize(self, key):
        """
        Read data and use the 'python_type' attribute to cast it back correctly.
        """
        try:
            dataset = self.file[key]
            type_name = dataset.attrs.get('python_type')

            if type_name is None:
                raise TypeError(f"Dataset for key '{key}' is missing 'python_type' attribute.")

            if type_name == 'NoneType':
                return None
            elif type_name in ('dict', 'list', 'tuple'):
                str_data = dataset.asstr()[()]
                return ast.literal_eval(str_data)
            elif type_name == 'str':
                return dataset.asstr()[()]
            else: 
                return dataset[()]

        except KeyError:
            raise KeyError(key)
        
    def del_item(self, key):
        try:
            del self.file[key]
        except KeyError:
            pass

    def keys(self):
        return list(self.file.keys())
    
    def destroy(self):
        self.file.close()
        os.remove(self.storage_path)    

class JSONBackend(DiskBackend):
    def __init__(self, storage_path):
        super().__init__(storage_path)
        os.makedirs(self.storage_path, exist_ok=True)

    def _file_path(self, key):
        return os.path.join(self.storage_path, f"{key}.json")

    def serialize(self, key, value):
        with open(self._file_path(key), "w") as f:
            json.dump(value, f)

    def deserialize(self, key):
        try:
            with open(self._file_path(key), "r") as f:
                return json.load(f)
        except FileNotFoundError:
            raise KeyError(key)
        
    def del_item(self, key):
        try:
            os.remove(self._file_path(key))
        except FileNotFoundError:
            pass

    def keys(self):
        return [os.path.splitext(f)[0] for f in os.listdir(self.storage_path)]
    
    def destroy(self):
        shutil.rmtree(self.storage_path)


from collections import OrderedDict, defaultdict
import random
from abc import abstractmethod

class ReplacementStrategy:
    def __init__(self, disk_backend, max_in_memory):
        self.memory = self.get_memory()
        self.disk_backend = disk_backend
        self.max_in_memory = max_in_memory

    @abstractmethod
    def get(self, key):
        pass

    @abstractmethod
    def put(self, key, value):
        pass
    
    @abstractmethod
    def get_memory(self):
        pass
        
    
    # --- NEW: Method for handling deletion from memory ---
    def delete(self, key):
        """Remove a key from internal memory structures."""
        if key in self.memory:
            del self.memory[key]

    def clear(self):
        """Clear internal memory structures."""
        self.memory.clear()

class RandomReplacement(ReplacementStrategy):
    def get(self, key):
        if key in self.memory:
            return self.memory[key]
        else:
            return self.disk_backend.deserialize(key)

    def put(self, key, value):
        self.memory[key] = value
        if len(self.memory) > self.max_in_memory:
            random_key = random.choice(list(self.memory.keys()))
            random_value = self.memory.pop(random_key)
            self.disk_backend.serialize(random_key, random_value)

    def get_memory(self):
        return defaultdict()
    
class FIFOReplacement(ReplacementStrategy):
    def get(self, key):
        if key in self.memory:
            return self.memory[key]
        else:
            return self.disk_backend.deserialize(key)
        
    def put(self, key, value):
        self.memory[key] = value
        if len(self.memory) > self.max_in_memory:
            oldest_key, oldest_value = self.memory.popitem(last=False)
            self.disk_backend.serialize(oldest_key, oldest_value)

    def get_memory(self):
        return OrderedDict()
    
class LIFOReplacement(ReplacementStrategy):
    def get(self, key):
        if key in self.memory:
            return self.memory[key]
        else:
            return self.disk_backend.deserialize(key)
        
    def put(self, key, value):
        self.memory[key] = value
        if len(self.memory) > self.max_in_memory:
            oldest_key, oldest_value = self.memory.popitem()
            self.disk_backend.serialize(oldest_key, oldest_value)

    def get_memory(self):
        return OrderedDict()
    
class LRUReplacement(ReplacementStrategy):
    def get(self, key):
        if key in self.memory:
            self.memory.move_to_end(key)
            return self.memory[key]
        else:
            value = self.disk_backend.deserialize(key)
            if value is not None:
                self.put(key, value)
            return value
        
    def put(self, key, value):
        self.memory[key] = value
        self.memory.move_to_end(key)
        if len(self.memory) > self.max_in_memory:
            oldest_key, oldest_value = self.memory.popitem(last=False)
            self.disk_backend.serialize(oldest_key, oldest_value)

    def get_memory(self):
        return OrderedDict()
    
class MRUReplacement(ReplacementStrategy):
    def get(self, key):
        if key in self.memory:
            self.memory.move_to_end(key)
            return self.memory[key]
        else:
            value = self.disk_backend.deserialize(key)
            if value is not None:
                self.put(key, value)
            return value
        
    def put(self, key, value):
        self.memory[key] = value
        self.memory.move_to_end(key)
        if len(self.memory) > self.max_in_memory:
            oldest_key, oldest_value = self.memory.popitem()
            self.disk_backend.serialize(oldest_key, oldest_value)

    def get_memory(self):
        return OrderedDict()
    
class LFUReplacement(ReplacementStrategy):
    def __init__(self, disk_backend, max_in_memory):
        super().__init__(disk_backend, max_in_memory)
        self.secondary_memory = defaultdict(int)

    def get(self, key):
        if key in self.memory:
            self.secondary_memory[key] += 1
            return self.memory[key]
        else:
            value = self.disk_backend.deserialize(key)
            self.put(key, value) # simplified: put handles loading
            return value
        
    def put(self, key, value):
        # Increment frequency if key exists, otherwise set to 1
        self.secondary_memory[key] = self.secondary_memory.get(key, 0) + 1
        self.memory[key] = value
        if len(self.memory) > self.max_in_memory:
            # Find the key with the minimum frequency to evict
            # Exclude the key we just added from eviction candidates
            eviction_candidates = {k: v for k, v in self.secondary_memory.items() if k != key}
            if eviction_candidates:
                min_key = min(eviction_candidates, key=eviction_candidates.get)
                min_value = self.memory.pop(min_key)
                self.disk_backend.serialize(min_key, min_value)
                del self.secondary_memory[min_key]

    def get_memory(self):
        return {} # Plain dict is fine here
    
    # --- OVERRIDE: Also clean up the frequency counter ---
    def delete(self, key):
        super().delete(key)
        if key in self.secondary_memory:
            del self.secondary_memory[key]

    def clear(self):
        super().clear()
        self.secondary_memory.clear()

# Similar changes for MFUReplacement...
class MFUReplacement(ReplacementStrategy):
    def __init__(self, disk_backend, max_in_memory):
        super().__init__(disk_backend, max_in_memory)
        self.secondary_memory = defaultdict(int)

    def get(self, key):
        if key in self.memory:
            self.secondary_memory[key] += 1
            return self.memory[key]
        else:
            value = self.disk_backend.deserialize(key)
            self.put(key, value)
            return value
        
    def put(self, key, value):
        self.secondary_memory[key] = self.secondary_memory.get(key, 0) + 1
        self.memory[key] = value
        if len(self.memory) > self.max_in_memory:
            eviction_candidates = {k: v for k, v in self.secondary_memory.items() if k != key}
            if eviction_candidates:
                max_key = max(eviction_candidates, key=eviction_candidates.get)
                max_value = self.memory.pop(max_key)
                self.disk_backend.serialize(max_key, max_value)
                del self.secondary_memory[max_key]

    def get_memory(self):
        return {}
    
    def delete(self, key):
        super().delete(key)
        if key in self.secondary_memory:
            del self.secondary_memory[key]

    def clear(self):
        super().clear()
        self.secondary_memory.clear()


    def __init__(self, disk_backend, max_in_memory):
        super().__init__(disk_backend, max_in_memory)
        self.secondary_memory = defaultdict(int)

    def get(self, key):
        if key in self.memory:
            self.secondary_memory[key] += 1
            return self.memory[key]
        else:
            value = self.disk_backend.deserialize(key)
            if value is not None:
                self.put(key, value)
            return value
        
    def put(self, key, value):
        self.secondary_memory[key] = 1
        self.memory[key] = value
        if len(self.memory) > self.max_in_memory:
            max_key = max(self.secondary_memory, key=self.secondary_memory.get)
            max_value = self.memory.pop(max_key)
            self.disk_backend.serialize(max_key, max_value)
            self.secondary_memory.pop(max_key)

    def get_memory(self):
        return OrderedDict()


# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information


import os
import sys
from pathlib import Path

# Add parent dir to known paths
p = Path(__file__).parents[2]
sys.path.insert(0, os.path.abspath(p))

project = "Effidict"
copyright = "2024, Helmholtz AI"
author = "Isra Mekki"
release = "0.0.1"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_book_theme"
html_static_path = ["_static"]
html_logo = "_static/logo_effidict.png"
html_favicon = "_static/logo_effidict.png"

html_title = "Effidict"
html_theme_options = {
    "logo_only": True,
    "repository_url": "https://github.com/HelmholtzAI-Consultants-Munich/EffiDict",
    "use_repository_button": True,
    "use_fullscreen_button": True,
    "use_issues_button": True,
}

autodoc_default_options = {
    "members": True,
    "undoc-members": True,
    "private-members": True,
}



from .effidict import EffiDict
from .disk_backend import DiskBackend, PickleBackend, Hdf5Backend, JSONBackend, SqliteBackend
from .replacement_strategies import ReplacementStrategy, RandomReplacement, FIFOReplacement, LIFOReplacement, LRUReplacement, MRUReplacement, LFUReplacement, MFUReplacement

__all__ = [
    "EffiDict",
    "DiskBackend",
    "PickleBackend",
    "Hdf5Backend",
    "JSONBackend",
    "SqliteBackend",
    "ReplacementStrategy",
    "RandomReplacement",
    "FIFOReplacement",
    "LIFOReplacement",
    "LRUReplacement",
    "MRUReplacement",
    "LFUReplacement",
    "MFUReplacement",
]


class EffiDict:
    def __init__(self, max_in_memory=100, disk_backend=None, replacement_strategy=None):
        self.max_in_memory = max_in_memory
        self.disk_backend = disk_backend
        self.replacement_strategy = replacement_strategy
        self.memory = replacement_strategy.memory

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        """Explicitly close the backend resources."""
        if self.disk_backend is not None:
            self.disk_backend.destroy()

    def __iter__(self):
        self._iter_keys = iter(self.keys())
        return self

    def __next__(self):
        return next(self._iter_keys)

    def __len__(self):
        return len(self.keys())

    def items(self):
        for key in self.keys():
            yield (key, self[key])

    def values(self):
        for key in self.keys():
            yield self[key]

    def __contains__(self, key):
        return key in self.memory or (self.disk_backend and key in self.disk_backend.keys())

    def pop(self, key, default=None):
        try:
            value = self.memory.pop(key)
        except KeyError:
            if key in self.keys():
                value = self[key]
                self.__delitem__(key)
            else:
                return default

        return value

    def clear(self):
        all_keys = self.keys()
        self.memory.clear()
        if self.disk_backend:
            for key in all_keys:
                self.disk_backend.del_item(key)
        # Also clear strategy-specific caches like LFU/MFU counts
        if hasattr(self.replacement_strategy, 'clear'):
            self.replacement_strategy.clear()


    def __del__(self):
        try:
            self.close()
        except Exception:
            pass

    def __eq__(self, other):
        if not isinstance(other, EffiDict):
            return False

        if len(self) != len(other):
            return False

        for key, value in self.items():
            if key not in other or other[key] != value:
                return False
        return True

    def keys(self):
        """Return a list of all unique keys known to the dictionary (memory + disk)."""
        mem_keys = set(self.memory.keys())
        disk_keys = set(self.disk_backend.keys() if self.disk_backend is not None else [])
        return list(mem_keys.union(disk_keys))

    def __getitem__(self, key):
        return self.replacement_strategy.get(key)

    def __setitem__(self, key, value):
        self.replacement_strategy.put(key, value)

    def __delitem__(self, key):
        # Check existence first to raise KeyError properly
        if key not in self:
            raise KeyError(key)
        # The strategy handles its own memory (and secondary_memory for LFU/MFU)
        self.replacement_strategy.delete(key)
        # Ensure it's also removed from the disk backend
        self.disk_backend.del_item(key)

    def load_from_dict(self, dictionary):
        self.disk_backend.load_from_dict(dictionary)


import pytest
import tempfile
import shutil
import os
import h5py

from effidict import (
    EffiDict,
    SqliteBackend, PickleBackend, Hdf5Backend, JSONBackend,
    RandomReplacement, FIFOReplacement, LIFOReplacement,
    LRUReplacement, MRUReplacement, LFUReplacement, MFUReplacement
)

BACKEND_CLASSES = [SqliteBackend, PickleBackend, JSONBackend]
if h5py:
    BACKEND_CLASSES.append(Hdf5Backend)

STRATEGY_CLASSES = [
    RandomReplacement, FIFOReplacement, LIFOReplacement,
    LRUReplacement, MRUReplacement, LFUReplacement, MFUReplacement,
]

TEST_COMBINATIONS = [
    pytest.param(backend, strategy, id=f"{backend.__name__}-{strategy.__name__}")
    for backend in BACKEND_CLASSES
    for strategy in STRATEGY_CLASSES
]

@pytest.fixture
def temp_dir_path():
    """Create a temporary directory for tests to use."""
    test_dir = tempfile.mkdtemp()
    yield test_dir  
    shutil.rmtree(test_dir) 


class TestEffiDictCRUD:

    @pytest.mark.parametrize("backend_cls, strategy_cls", TEST_COMBINATIONS)
    def test_within_capacity(self, temp_dir_path, backend_cls, strategy_cls):
        """
        Tests CRUD operations when the number of items is less than max_in_memory.
        """
        max_items = 5
        storage_path = os.path.join(temp_dir_path, backend_cls.__name__)
        
        disk_backend = backend_cls(storage_path)
        replacement_strategy = strategy_cls(disk_backend=disk_backend, max_in_memory=max_items)
        
        with EffiDict(
            max_in_memory=max_items,
            disk_backend=disk_backend,
            replacement_strategy=replacement_strategy
        ) as effidict:
            # CREATE
            effidict['a'] = 1
            effidict['b'] = {'data': [2, 3]}
            assert len(effidict) == 2
            assert len(effidict.memory) == 2
            assert len(effidict.disk_backend.keys()) == 0

            # READ
            assert effidict['a'] == 1
            assert effidict['b'] == {'data': [2, 3]}
            assert 'a' in effidict

            # UPDATE
            effidict['a'] = 100
            assert effidict['a'] == 100

            # DELETE
            del effidict['a']
            assert len(effidict) == 1
            value = effidict.pop('b')
            assert value == {'data': [2, 3]}
            assert len(effidict) == 0

    @pytest.mark.parametrize("backend_cls, strategy_cls", TEST_COMBINATIONS)
    def test_disk_is_used(self, temp_dir_path, backend_cls, strategy_cls):
        """
        Tests CRUD operations when items exceed max_in_memory, forcing disk usage.
        """
        max_items = 3
        num_to_add = 5
        storage_path = os.path.join(temp_dir_path, backend_cls.__name__)
        
        disk_backend = backend_cls(storage_path)
        replacement_strategy = strategy_cls(disk_backend=disk_backend, max_in_memory=max_items)

        with EffiDict(
            max_in_memory=max_items,
            disk_backend=disk_backend,
            replacement_strategy=replacement_strategy
        ) as effidict:
            items = {f'key{i}': f'value{i}' for i in range(num_to_add)}
            for key, value in items.items():
                effidict[key] = value

            assert len(effidict) == num_to_add
            assert len(effidict.memory) == max_items
            assert len(effidict.disk_backend.keys()) == num_to_add - max_items

            # READ from disk
            disk_keys = set(effidict.disk_backend.keys())
            key_on_disk = list(disk_keys)[0]
            assert effidict[key_on_disk] == items[key_on_disk]
            
            # UPDATE on disk
            effidict[key_on_disk] = "updated_value"
            assert effidict[key_on_disk] == "updated_value"

            # DELETE from disk
            del effidict[key_on_disk]
            assert len(effidict) == num_to_add - 1
            assert key_on_disk not in effidict

            # Test deleting a non-existent key
            with pytest.raises(KeyError):
                del effidict['non_existent_key']